package millikanexperimentapp;

import java.awt.Graphics;
import org.opensourcephysics.display.Drawable;
import org.opensourcephysics.display.DrawingPanel;
import java.awt.Color;

/**
 * <p>Title: Millikan experiment</p>
 * <p>Description: This is a drawable set of oil drops in the Millikan experiment.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Gymnazium arm. gen. L. Svobodu</p>
 * @author Jakub Dubas, Rudolf Tomori, and Slavomir Tuleja
 * @version 2.7 from February 15, 2005
 */

public class OilDrops
    implements Drawable {
  int maxNumDrops = 100;
  int numWalkers = 0;
  int voltage;
  double dt = 0.01;

  private final double DENSITY_OIL = 875.3; //in units kg/m^3
  private final double DENSITY_AIR = 1.29; // in units kg/m^3
  private final double ETA = 7.25e-6; // viscosity of air in units N.s/m^2
  private final double DISTANCE_OF_PLATES = 6e-3; //in meters
  private final double kB = 1.38e-23; //Boltzmann constant in J/K
  private final double G = 9.8; // gravitational acceleration in units m/s^2

  private double[] alpha = new double[maxNumDrops]; //friction constant in Langevin equation
  private double[] m = new double[maxNumDrops]; //mass of the drop in kilograms
  private double Temp = 293; //temperature of air in Kelvins
  private double[] r = new double[maxNumDrops]; //radius of oil drop in meters;
  private double[] charge = new double[maxNumDrops]; //charge on the drop in units coulombs
  private double[] x = new double[maxNumDrops]; //x-positions of drops
  private double[] y = new double[maxNumDrops]; //y-positions of drops
  private double[] v = new double[maxNumDrops]; //terminal velocities of drops

  /**
   * Removes drops that have reached the capacitor plates.
   */
  public void removeGarbageDrops() {
    //look for drop which is on charged plate
    for (int i = 0; i < numWalkers; i++) {
      if ( (y[i] < -3e-3) || (y[i] > 3e-3)) {
        //remove drop which is on charged plate
        for (int j = i; j < numWalkers - 1; j++) {
          x[j] = x[j + 1];
          y[j] = y[j + 1];
          alpha[j] = alpha[j + 1];
          m[j] = m[j + 1];
          charge[j] = charge[j + 1];
          r[j] = r[j + 1];
          v[j] = v[j + 1];
        }
        numWalkers--;
      }
    }
  }

  /**
   * Generates two normally distributed random numbers and adds them to the position
   * of the i-th drop. This simulates the displacement resulting in one time step
   * from the Brownian motion of the drop.
   * @param i int index of the drop
   */
  public void generateRandomCorrection(int i) {
    float x1, x2, w;

    double meanDisplacementSquared = 2 * kB * Temp / alpha[i] * 0.01;

    //We generate two random numbers from normal distribution N(0,<x^2>)...
    do {
      x1 = (float) (2.0 * Math.random() - 1.0);
      x2 = (float) (2.0 * Math.random() - 1.0);
      w = x1 * x1 + x2 * x2;
    }
    while (w >= 1.0);
    w = (float) (Math.sqrt(meanDisplacementSquared) *
                 Math.sqrt( ( -2.0 * Math.log(w)) / w));
    x[i] += x1 * w;
    y[i] += x2 * w;
  }

  /**
   * If the user changes value of the voltage on the slider in the
   * MillikanExperimentApp we have to adjust the terminal velocities of the
   * drops accordingly. This is done here.
   * @param sl_voltage int integer value of voltage on the slider in MillikanExperimentApp.
   */
  public void setVelocity(int sl_voltage) {
    voltage = sl_voltage;
    for (int i = 0; i < numWalkers; i++)
      v[i] = ( +4.0 / 3.0 * Math.PI * r[i] * r[i] * r[i] *
              (DENSITY_OIL - DENSITY_AIR) * G -
              voltage / DISTANCE_OF_PLATES * charge[i]) / alpha[i];

  }

  /**
   * Shifts all the drops to their new positions after each time step dt.
   *
   * @param rayon boolean true if the ionizing radiation is on.
   * @param voltageon boolean true if the voltage os on.
   */
  public void doStep(boolean rayon, boolean voltageon) {
    removeGarbageDrops();

    //move drop
    for (int i = 0; i < numWalkers; i++) {
      y[i] += v[i] * dt;
      generateRandomCorrection(i);

      //change charge of drop if drop acrossed ray
      if (rayon) {
        if (Math.random() < 0.01) {
          //generate random number of elementary chrges the drop acquires after it
          //is ionized by a ray; n=-2,-1,..1,2
          int n = Math.round( (float) ( -2 + 4 * Math.random()));
          charge[i] += (1.602e-19) * n;
        }
      }

    }
    //adjust terminal velocities of the drops if the voltage is on
    if (voltageon) setVelocity(voltage);
  }

  /**
   * Injects (creates) a new group of 10 randomly charged drops.
   */
  public void inject() {
    float x1, x2, w;
    int oldNumWalkers;
    oldNumWalkers = numWalkers;
    for (int i = oldNumWalkers; i < oldNumWalkers + 10; i++) {

      // inject new drops only if number of drops is smaller than maxNumDrops
      if (numWalkers >= maxNumDrops)
        break;

      //Set random position of the i-th drop near the bottom of the viewfield
      //of the microscope. Positions are evenly
      //distributed between x=-0.75e-3 and x=0.75e-3.
      x[i] = -0.75e-3 + 2 * 0.75e-3 * Math.random();
      y[i] = -0.80e-3;
      //Set random number of elementary charges on the drop; n=-4,-3,...,3,4
      charge[i] = (Math.round(8 * Math.random()) - 4) * 1.602e-19;

      //set the random radius
      //generate random number from normal distribution N(0, (0.1e-6)^2 )
      do {
        x1 = (float) (2.0 * Math.random() - 1.0);
        x2 = (float) (2.0 * Math.random() - 1.0);
        w = x1 * x1 + x2 * x2;
      }
      while (w >= 1.0);
      w = (float) (0.1e-6 * Math.sqrt( ( -2.0 * Math.log(w)) / w));

      //set values for new drop
      r[i] = (x1 * w) + 0.5e-6;

      alpha[i] = 6 * Math.PI * ETA * r[i];
      m[i] = 4.0 / 3.0 * Math.PI * r[i] * r[i] * r[i] * DENSITY_OIL;
      numWalkers++;
    }
  }

  public void draw(DrawingPanel panel, Graphics g) {

    //draw lines
    g.setColor(new Color(200, 200, 130));
    for (int i = -5; i <= 5; i++) {
      g.drawLine(panel.xToPix( -0.7e-3), panel.yToPix(i * 0.1e-3),
                 panel.xToPix(0.7e-3), panel.yToPix(i * 0.1e-3));
    }

    //draw drops
    g.setColor(new Color(200, 200, 230));
    for (int i = 0; i < numWalkers; i++)
      g.fillOval(panel.xToPix(x[i]) - 2, panel.yToPix(y[i]) - 2, 4, 4);
  }

}
